use rand::Rng;
use std::collections::HashMap;

#[derive(Debug, Clone, serde::Deserialize)]
pub struct Config {
    pub interfaces: HashMap<String, InterfaceConfig>,
}

#[derive(Debug, Copy, Clone, serde::Deserialize)]
#[serde(default)]
pub struct InterfaceConfig {
    pub mcast_hello_interval: u16,
    pub ucast_hello_interval: u16,
    pub link_cost: u16,
    pub ihu_interval: u16,
    pub update_interval: u16,
    pub hold_time: u16,
    pub expiry_time: u16,
    pub timeout: u16,
    pub urgent_timeout: f32,
    pub source_gc: u16,
    pub router_id: u64,
}

#[derive(Clone)]
pub struct Advertise {
    pub prefix: Vec<cidr::IpInet>,
}

impl Advertise {
    pub fn from_vec_str(vec: Vec<String>) -> anyhow::Result<Self> {
        let mut prefix = Vec::new();
        for s in vec {
            let Some((addr, mask)) = s.split_once('/') else { anyhow::bail!(format!("Invalid address: {}", s)) };
            prefix.push(cidr::IpInet::new(addr.parse()?, mask.parse()?)?);
        }
        Ok(Self { prefix })
    }
}

#[derive(Debug, serde::Deserialize)]
pub struct Conf {
    prefix: Vec<String>,
    pub interface: HashMap<String, InterfaceConfig>,
}

impl Default for InterfaceConfig {
    fn default() -> Self {
        Self {
            mcast_hello_interval: 400,
            ucast_hello_interval: 0,
            link_cost: 97,
            ihu_interval: 3 * 400,
            update_interval: 4 * 400,
            hold_time: (3.5 * 3.0 * 400.0) as u16,
            expiry_time: (3.5 * 4.0 * 400.0) as u16,
            timeout: 200,
            urgent_timeout: 0.2,
            source_gc: 180,
            router_id: rand::thread_rng().gen_range(1..u16::MAX) as u64,
        }
    }
}

impl Default for Advertise {
    fn default() -> Self {
        Self { prefix: Vec::new() }
    }
}

impl Conf {
    pub async fn load(path: String) -> anyhow::Result<Self> {
        let files = tokio::fs::read_to_string(path).await?;
        let conf = toml::from_str(&files)?;
        Ok(conf)
    }

    pub fn get(self) -> anyhow::Result<(Config, Advertise)> {
        let config = Config {
            interfaces: self.interface,
        };
        let advertise = Advertise::from_vec_str(self.prefix)?;
        Ok((config, advertise))
    }
}
