use crate::config::Advertise;
use crate::config::InterfaceConfig;
use crate::nht::Nht;
use crate::tables::neighbor::NeighborTable;
use crate::threads::Worker;
use crate::tlvs::{Hello, HelloFlags, IntoPacket, NextHop, RouterId, Update, UpdateFlags, IHU};
use nix::net::if_::if_nametoindex;
use rand::Rng;
use std::collections::HashMap;
use std::sync::atomic::{AtomicU16, Ordering};
use std::sync::Arc;
use tokio::sync::Mutex;
use tokio::task::JoinHandle;

// List of active interface
// K: interface ID
// V: interface entry
pub struct InterfaceTable {
    active_interfaces: HashMap<u32, Interface>,
}

// Timers:
// - Hello timer
// - Update timer
pub struct Interface {
    pub mcast_seqno: Arc<AtomicU16>,
    pub worker: Worker,
    pub hello_timer: JoinHandle<()>,
    pub ihu_timer: JoinHandle<()>,
    pub update_timer: JoinHandle<()>,
}

impl InterfaceTable {
    pub fn new() -> Self {
        Self {
            active_interfaces: HashMap::new(),
        }
    }

    pub async fn add_by_name(
        &mut self,
        name: &str,
        conf: InterfaceConfig,
        advertise: Advertise,
        neighbor_table: Arc<Mutex<NeighborTable>>,
    ) -> anyhow::Result<()> {
        let id = if_nametoindex(name)? as u32;
        let worker = Worker::new(id).await?;
        let interface = Interface::new(worker, conf, advertise, neighbor_table, id).await?;

        self.active_interfaces.insert(id, interface);
        Ok(())
    }

    /*pub fn get_interface_index(&self, iface_id: u32) -> Option<&Interface> {
        self.active_interfaces.get(&iface_id)
    }*/

    pub fn get_interface<'a>(&'a mut self, index: u32) -> Option<&'a mut Interface> {
        self.active_interfaces.get_mut(&index)
    }

    /*pub fn del_interface(&mut self, iface_id: u32) -> Option<Interface> {
        let iface = self.active_interfaces.remove(&iface_id);
        if let Some(iface) = iface {
            iface.hello_timer.abort();
            return Some(iface);
        }
        None
    }*/
}

impl Interface {
    // Thread connection to worker thread to send scheduled mcast hellos, IHUs and Updates
    pub async fn new(
        worker: Worker,
        conf: InterfaceConfig,
        advertise: Advertise,
        neighbor_table: Arc<Mutex<NeighborTable>>,
        id: u32,
    ) -> anyhow::Result<Self> {
        let mcast_seqno = Arc::new(AtomicU16::new(0));

        let hello_worker = worker.clone();
        let hello_seqno = mcast_seqno.clone();
        let hello_interval = conf.mcast_hello_interval as u64;
        let hello_timer = tokio::spawn(async move {
            Self::hello_timer(hello_worker, hello_seqno, hello_interval).await
        });

        let ihu_worker = worker.clone();
        let ihu_interval = conf.ihu_interval as u64;
        let ihu_timer = tokio::spawn(async move {
            Self::ihu_timer(ihu_worker, neighbor_table, ihu_interval, id).await
        });

        let update_worker = worker.clone();
        let update_interval = conf.update_interval as u64;
        let update_rxcost = conf.link_cost;
        let update_router_id = conf.router_id;
        let update_timer = tokio::spawn(async move {
            Self::update_timer(
                update_worker,
                advertise,
                id,
                update_router_id,
                update_rxcost,
                update_interval,
            )
            .await
        });

        if update_timer.is_finished() {
            anyhow::bail!("Couldn't get ifname of interface");
        }

        Ok(Self {
            mcast_seqno,
            worker,
            hello_timer,
            ihu_timer,
            update_timer,
        })
    }

    pub async fn hello_timer(mut worker: Worker, mcast_seqno: Arc<AtomicU16>, interval: u64) {
        loop {
            tracing::debug!("Sending Hello");
            let seqno = mcast_seqno.load(Ordering::Relaxed);
            let hello = Hello {
                flags: HelloFlags::Multicast,
                interval: interval as u16,
                seqno,
            };
            let jitter = rand::thread_rng().gen_range(0..(interval / 2));

            worker.send_packet(hello.into_packet(), false, jitter).await;
            mcast_seqno.store(seqno.wrapping_add(1), Ordering::Release);

            let sleep = std::time::Duration::from_millis((interval + jitter) * 10);
            tokio::time::sleep(sleep).await;
        }
    }

    pub async fn ihu_timer(
        mut worker: Worker,
        neighbor_table: Arc<Mutex<NeighborTable>>,
        interval: u64,
        id: u32,
    ) {
        loop {
            let jitter = rand::thread_rng().gen_range(0..(interval / 2));
            let timeout = (interval as u64 + jitter as u64) * 10;
            tokio::time::sleep(tokio::time::Duration::from_millis(timeout)).await;

            let mut neighbor_table = neighbor_table.lock().await;
            for (index, neighbor) in neighbor_table.get_neighbors() {
                if index.iface == id {
                    tracing::debug!("Sending IHU to {} on iface {}", index.address, index.iface);
                    let ihu = IHU::new(neighbor.rx_cost, interval as u16, index.address);
                    worker
                        .send_packet(ihu.into_packet(), false, jitter as u64)
                        .await;
                }
            }
        }
    }

    pub async fn update_timer(
        mut worker: Worker,
        advertise: Advertise,
        id: u32,
        router_id: u64,
        rx_cost: u16,
        interval: u64,
    ) {
        let ifname = Nht::get_ifname_by_index(id);
        if let Err(e) = ifname {
            tracing::error!("Error getting interface name: {e}");
            return;
        }

        let mut seqno = 0;
        let ifname = ifname.unwrap();
        loop {
            // Router-id
            // v6 prefix
            // next hop
            // v4 prefix
            // neighbor router-id
            // v6 prefix
            // v4 prefix
            tracing::debug!("Sending Update");
            let jitter = rand::thread_rng().gen_range(0..(interval / 2));
            let timeout = (interval as u64 + jitter as u64) * 10;

            let router_id = RouterId { router_id };
            worker
                .send_packet(router_id.into_packet(), false, jitter)
                .await;

            for prefix in &advertise.prefix {
                let ae = if prefix.is_ipv6() {
                    2
                } else {
                    let addresses = Nht::get_ifaddrs(ifname.clone());
                    if let Err(e) = addresses {
                        tracing::error!("Error getting addresses of interface: {e}");
                        continue;
                    }

                    if let Some(next_hop) = Nht::get_primary_v4(addresses.unwrap()) {
                        let next_hop = NextHop { ae: 1, next_hop };
                        worker
                            .send_packet(next_hop.into_packet(), false, jitter)
                            .await;
                    } else {
                        tracing::warn!("No IPv4 addresses on interface, ignoring");
                    }

                    1
                };
                let update = Update {
                    ae,
                    plen: prefix.network_length(),
                    flags: UpdateFlags::P,
                    omitted: 0,
                    interval: interval as u16,
                    seqno,
                    metric: rx_cost,
                    prefix: prefix.first_address(),
                };

                worker
                    .send_packet(update.into_packet(), false, jitter)
                    .await;
            }

            seqno = seqno.wrapping_add(1);
            tokio::time::sleep(std::time::Duration::from_millis(timeout)).await;
        }
    }
}
