use crate::tlvs::{Hello, HelloFlags, IHU};
use std::sync::atomic::AtomicU16;
use std::{collections::HashMap, net::IpAddr, sync::Arc};
use tokio::sync::mpsc::{unbounded_channel, UnboundedReceiver, UnboundedSender};
use tokio::task::JoinHandle;

pub struct NeighborTable {
    active_interfaces: HashMap<NeighborTableIndex, NeighborTableEntry>,
}

#[derive(PartialEq, Eq, Hash, Clone, Copy)]
pub struct NeighborTableIndex {
    pub iface: u32,
    pub address: IpAddr,
}

pub struct NeighborTableEntry {
    ucast_conn: UnboundedSender<Hello>,
    mcast_conn: UnboundedSender<Hello>,
    ihu_conn: UnboundedSender<IHU>,
    pub ucast_timer: JoinHandle<()>,
    pub mcast_timer: JoinHandle<()>,
    pub ihu_timer: JoinHandle<()>,
    pub tx_cost: Arc<AtomicU16>,
    pub rx_cost: u16,
    pub ucast_seqno_out: u16,
}

impl NeighborTable {
    pub fn new() -> Self {
        Self {
            active_interfaces: HashMap::new(),
        }
    }

    pub fn add_neighbor(&mut self, index: NeighborTableIndex, entry: NeighborTableEntry) {
        self.active_interfaces.insert(index, entry);
    }

    pub fn get_neighbor<'a>(
        &'a mut self,
        index: &NeighborTableIndex,
    ) -> Option<&'a mut NeighborTableEntry> {
        self.active_interfaces.get_mut(index)
    }

    pub fn get_neighbors<'a>(
        &'a mut self,
    ) -> Vec<(&'a NeighborTableIndex, &'a NeighborTableEntry)> {
        let mut ret = Vec::new();
        for (index, neighbor) in &self.active_interfaces {
            ret.push((index, neighbor));
        }
        ret
    }

    pub fn del_neighbor(&mut self, index: &NeighborTableIndex) {
        let neighbor = self.active_interfaces.remove(index);
        if let Some(neighbor) = neighbor {
            neighbor.ucast_timer.abort();
            neighbor.mcast_timer.abort();
            neighbor.ihu_timer.abort();
        }
    }
}

impl NeighborTableIndex {
    // TODO: Don't just assume we are using IPv6 link locals
    pub fn parse(str: String) -> anyhow::Result<Self> {
        let split = str.split_once('%');

        if let Some((ip_addr, iface_index)) = split {
            let ip_addr = ip_addr.strip_prefix('[');
            let iface_index = iface_index.split_once(':');
            if iface_index.is_none() {
                anyhow::bail!("Invalid socket address");
            }
            let iface_index = iface_index.unwrap().0;
            let iface_index = iface_index.strip_suffix(']');

            if let (Some(ip_addr), Some(iface_index)) = (ip_addr, iface_index) {
                let address = ip_addr.parse()?;
                let iface = iface_index.parse()?;

                Ok(Self { address, iface })
            } else {
                anyhow::bail!("Failed to parse IPv6 link local address");
            }
        } else {
            anyhow::bail!("Failed to parse IPv6 link local address");
        }
    }
}

impl NeighborTableEntry {
    pub async fn new() -> Self {
        let rx_cost = 1;
        let tx_cost = Arc::new(AtomicU16::new(0xFFFF));
        let ihu_tx_cost = tx_cost.clone();

        let (ucast_conn, ucast_rx) = unbounded_channel();
        let (mcast_conn, mcast_rx) = unbounded_channel();
        let (ihu_conn, ihu_rx) = unbounded_channel();

        let mcast_timer = tokio::spawn(async { Self::mcast_hello_timer(mcast_rx).await });
        let ucast_timer = tokio::spawn(async { Self::ucast_hello_timer(ucast_rx).await });
        let ihu_timer = tokio::spawn(async { Self::ihu_timer(ihu_rx, ihu_tx_cost).await });

        Self {
            tx_cost,
            rx_cost,
            ucast_seqno_out: 0,
            ucast_conn,
            mcast_conn,
            ihu_conn,
            mcast_timer,
            ucast_timer,
            ihu_timer,
        }
    }

    pub async fn add_hello(&mut self, hello: Hello) -> anyhow::Result<()> {
        match hello.flags {
            HelloFlags::Multicast => self.mcast_conn.send(hello)?,
            HelloFlags::Unicast => self.ucast_conn.send(hello)?,
        }
        Ok(())
    }

    pub async fn add_ihu(&mut self, ihu: IHU) -> anyhow::Result<()> {
        self.ihu_conn.send(ihu)?;
        Ok(())
    }

    fn check_queue(queue: &Vec<bool>) -> bool {
        if queue.len() < 14 {
            return true;
        }
        for entry in queue {
            if *entry {
                return true;
            }
        }
        false
    }

    async fn ihu_timer(mut rx: UnboundedReceiver<IHU>, rx_cost: Arc<AtomicU16>) {
        let mut duration = 0;

        if let Some(ihu) = rx.recv().await {
            duration = (ihu.interval + (ihu.interval / 2)) as u64;
        }

        let duration = tokio::time::Duration::from_millis(duration * 10);
        loop {
            let timeout = tokio::time::sleep(duration);
            tokio::select! {
                _ = timeout => {
                    tracing::debug!("IHU timer timed out");
                    rx_cost.store(0xFFFF, std::sync::atomic::Ordering::Relaxed);
                }
                Some(ihu) = rx.recv() => rx_cost.store(ihu.rxcost, std::sync::atomic::Ordering::Relaxed),
            }
        }
    }

    crate::add_timer!(ucast_hello_timer);
    crate::add_timer!(mcast_hello_timer);
}

#[macro_export]
macro_rules! add_timer {
    ($name:tt) => {
        async fn $name(mut rx: UnboundedReceiver<Hello>) {
            let mut duration = 0;
            let mut hello_hist = Vec::new();
            let mut expected_seqno = 0;

            if let Some(hello) = rx.recv().await {
                expected_seqno = hello.seqno;
                duration = (hello.interval as f32 * 1.5) as u64;
            }

            loop {
                let timeout = tokio::time::sleep(tokio::time::Duration::from_millis(duration * 10));
                if hello_hist.len() >= 14 {
                    let _ = hello_hist.remove(0);
                }

                tokio::select! {
                    _ = timeout => {
                        tracing::debug!("Hello timer timed out");
                        hello_hist.push(false);
                    }
                    Some(hello) = rx.recv() => {
                        if hello_hist.len() == 0 {
                            expected_seqno = hello.seqno;
                        } else if expected_seqno != hello.seqno {
                            expected_seqno = hello.seqno;
                            tracing::debug!("Neighbor send seqno {}, expected {expected_seqno}", hello.seqno);
                            hello_hist.push(false);
                        } else {
                            hello_hist.push(true);
                        }
                        duration = (hello.interval as f32 * 1.5) as u64;
                        expected_seqno = expected_seqno.wrapping_add(1);
                    },
                }

                tracing::trace!("Hello history: {hello_hist:?}");
                if !Self::check_queue(&hello_hist) {
                    tracing::debug!(
                        "Didn't receive for the last {}, dropping neighbor",
                        15 * duration
                    );
                    return;
                }
            }
        }
    };
}
