use std::{collections::HashMap, net::IpAddr, time::Duration};
use tokio::sync::mpsc::{unbounded_channel, UnboundedReceiver, UnboundedSender};
use tokio::task::JoinHandle;

// SouceTable
// K: SourceTableIndex
// V: (seqno, metric)
pub struct SourceTable {
    feasibility_distances: HashMap<SourceTableIndex, SourceTableEntry>,
}

// Index for the source table
// - Prefix
// - Plen
// - RouterId
// Timers:
// - Garbage collection timer
#[derive(Eq, Hash, PartialEq)]
pub struct SourceTableIndex {
    prefix: IpAddr,
    plen: u8,
    router_id: u64,
}

pub struct SourceTableEntry {
    pub seqno: u16,
    pub metric: u16,
    pub gc_conn: UnboundedSender<()>,
    pub gc_timer: JoinHandle<()>,
}

impl SourceTable {
    pub fn new() -> Self {
        let feasibility_distances = HashMap::new();
        Self {
            feasibility_distances,
        }
    }

    pub async fn update(
        &mut self,
        index: SourceTableIndex,
        seqno: u16,
        metric: u16,
    ) -> anyhow::Result<()> {
        if let Some(entry) = self.feasibility_distances.get_mut(&index) {
            if seqno > entry.seqno {
                entry.seqno = seqno;
                entry.metric = metric;
                entry.gc_conn.send(())?;
            } else if seqno == entry.seqno && metric > entry.seqno {
                entry.metric = metric;
                if metric != 0xFFFF {
                    entry.gc_conn.send(())?;
                }
            }
        } else {
            let entry = SourceTableEntry::new(seqno, metric).await;
            self.feasibility_distances.insert(index, entry);
        }

        Ok(())
    }
}

impl SourceTableIndex {
    pub fn new(prefix: IpAddr, plen: u8, router_id: u64) -> Self {
        Self {
            prefix,
            plen,
            router_id,
        }
    }
}

impl SourceTableEntry {
    pub async fn new(seqno: u16, metric: u16) -> Self {
        let (gc_conn, gc_rx) = unbounded_channel();
        let gc_timer = tokio::task::spawn(async { Self::gc_timer(1000, gc_rx).await });

        Self {
            seqno,
            metric,
            gc_conn,
            gc_timer,
        }
    }

    async fn gc_timer(gc_timeout: u64, mut rx: UnboundedReceiver<()>) {
        let duration = Duration::from_millis(gc_timeout * 10);
        loop {
            let timeout = tokio::time::sleep(duration);
            tokio::select! {
                _ = timeout => return,
                _ = rx.recv() => continue,
            }
        }
    }
}
