use crate::{
    config::{Advertise, Config},
    tables::neighbor::NeighborTable,
    tlvs::{Hello, HelloFlags, IntoPacket, Packet, IHU},
    transport::Socket,
};
use rand::Rng;
use std::sync::{
    atomic::{AtomicBool, Ordering},
    Arc,
};
use tokio::sync::Mutex;

#[derive(Clone)]
pub struct Worker {
    socket: Socket,
    send_lock: Arc<AtomicBool>,
}

impl Worker {
    pub async fn new(iface_index: u32) -> anyhow::Result<Self> {
        Ok(Self {
            socket: Socket::new(iface_index).await?,
            send_lock: Arc::new(AtomicBool::new(false)),
        })
    }

    pub async fn handle(
        &mut self,
        neighbor_table: Arc<Mutex<NeighborTable>>,
    ) -> anyhow::Result<()> {
        let recv_neighbor_table = neighbor_table.clone();
        let recv_socket = self.socket.clone();
        // let tx_neighbor_table = neighbor_table.clone();
        // let tx_socket = socket.clone();

        tokio::task::spawn(
            async move { Self::receive_loop(recv_socket, recv_neighbor_table).await },
        );
        // tokio::task::spawn(async move { Self::send_loop(tx_socket, tx_neighbor_table, config, advertise).await });

        loop {
            tokio::time::sleep(std::time::Duration::from_millis(1000)).await;
        }
    }

    // Receiving and replying
    async fn receive_loop(socket: Socket, neighbor_table: Arc<Mutex<NeighborTable>>) {
        loop {
            if let Err(e) = socket.recv(neighbor_table.clone()).await {
                tracing::error!("Error in receive function: {e}");
            }
        }
    }

    pub async fn send_packet(&mut self, packet: Packet, urgent: bool, jitter: u64) {
        self.socket.send(packet).await;
        if !urgent {
            if self.send_lock.load(Ordering::Acquire) {
                return;
            }

            self.send_lock.store(true, Ordering::Relaxed);
            tokio::time::sleep(std::time::Duration::from_millis(jitter * 10)).await;
            if !self.send_lock.load(Ordering::Acquire) {
                return;
            }
        }

        self.send_lock.store(false, Ordering::Release);
        if let Err(e) = self.socket.tx().await {
            tracing::error!("Error signalling send: {e}");
        }
    }

    /*
    async fn scheduled_hellos(&mut self, interval: u64) {
        let mut seqno = 0;
        loop {
            tracing::debug!("Sending Hello");
            let hello = Hello { flags: HelloFlags::Multicast, interval: interval as u16, seqno };
            let jitter = rand::thread_rng().gen_range(0..(interval / 2));
            self.send_packet(hello.into_packet(), false, jitter).await;
            seqno = seqno.wrapping_add(1);
            let sleep = std::time::Duration::from_millis((interval + jitter) * 10);
            tokio::time::sleep(sleep).await;
        }
    }

    // IHU stuff
    async fn scheduled_ihu(&mut self, interval: u16, neighbor_table: Arc<Mutex<NeighborTable>>) {
        loop {
            let jitter = rand::thread_rng().gen_range(0..(interval / 2));
            let timeout = (interval as u64 + jitter as u64) * 10;
            tokio::time::sleep(tokio::time::Duration::from_millis(timeout)).await;
            let mut neighbor_table = neighbor_table.lock().await;
            let mut del = Vec::new();
            let neighbors = neighbor_table.get_neighbors();

            for (index, neighbor) in neighbors.clone() {
                tracing::trace!("Sending IHU to {} on {}", index.address, index.iface);
                if neighbor.ucast_timer.is_finished() || neighbor.mcast_timer.is_finished() {
                    del.push(*index);
                }

                let ihu = IHU::new(neighbor.rx_cost, interval, index.address);
                self.send_packet(ihu.into_packet(), false, jitter as u64).await;
            }

            for index in del {
                neighbor_table.del_neighbor(&index);
            }
        }
    }

    async fn scheduled_updates(&mut self, interval: u64, prefixes: Advertise) {
        let sleep = std::time::Duration::from_millis(interval * 10);
        let mut seqno = rand::thread_rng().gen_range(0..u16::MAX);
        loop {
            tracing::trace!("Sending Update");
            for prefix in &prefixes.prefix {
                if prefix.is_ipv4() {
                    // NHT, update
                } else {
                    // Update
                }
            }

            seqno = seqno.wrapping_add(1);
            tokio::time::sleep(sleep).await;
        }
    }
    */
}

/*use std::net::IpAddr;

use crate::{
    config::{Advertise, Config},
    tables::neighbor::NeighborTable,
    tlvs::{Hello, IntoPacket, NextHop, Packet, RouterId, Update, UpdateFlags, IHU},
    transport::Socket,
};
use tokio::{
    sync::mpsc::{unbounded_channel, UnboundedReceiver, UnboundedSender},
    sync::{Arc, Mutex},
};
use rand::Rng;

#[derive(Clone)]
pub struct Threadpool {
    task_conn_rx: UnboundedReceiver<Packet>,
    task_conn_tx: UnboundedSender<Packet>,
    signal_conn_rx: UnboundedReceiver<()>,
    signal_conn_tx: UnboundedSender<()>,
}

impl Threadpool {
    pub fn new() -> Self {
        let (task_conn_tx, task_conn_rx) = unbounded();
        let (signal_conn_tx, signal_conn_rx) = unbounded();
        Self {
            task_conn_rx,
            task_conn_tx,
            signal_conn_rx,
            signal_conn_tx,
        }
    }

    pub async fn sock_handler(self, sock: Socket, neighbors: Arc<Mutex<NeighborTable>>) {
        async_std::task::spawn(async move {
            loop {
                if let Err(e) = sock.recv(neighbors.clone()).await {
                    log::error!("Error receiving from UdpSocket: {e}");
                }
            }
        });
    }

    pub async fn send_handler(self, mut sock: Socket) {
        async_std::task::spawn(async move {
            loop {
                let packet = self.task_conn_rx.recv().await;
                if let Err(e) = packet {
                    log::error!("Error receiving from thread connection: {e}");
                } else {
                    sock.send(packet.unwrap()).await;
                }

                let should_send = self.signal_conn_rx.try_recv();
                if let Err(e) = should_send {
                    if !e.is_empty() {
                        log::error!("Error receiving from thread connection: {e}");
                    }
                } else {
                    if let Err(e) = sock.tx().await {
                        log::error!("Error sending packet(s): {e}");
                    }
                }
            }
        });
    }

    async fn sender(self, is_urgent: bool, conf: Config) {
        async_std::task::spawn(async move {
            if !is_urgent {
                let delay = rand::thread_rng().gen_range(0..(conf.mcast_hello_interval / 2));
                async_std::task::sleep(std::time::Duration::from_millis(delay as u64 * 10)).await;
            }

            if let Err(e) = self.signal_conn_tx.send(()).await {
                log::error!("Error sending to thread connection: {e}");
            }
        });
    }

    pub async fn hello_timer(self, conf: Config) {
        let pow = 2_u32.pow(16);
        let mut seqno = (rand::thread_rng().gen_range(0..u16::MAX) as u32).rem_euclid(pow) as u16;
        async_std::task::spawn(async move {
            loop {
                let hello = Hello {
                    flags: crate::tlvs::HelloFlags::Multicast,
                    interval: conf.mcast_hello_interval,
                    seqno,
                };
                if let Err(e) = self.task_conn_tx.send(hello.into_packet()).await {
                    log::error!("Error sending to thread connection: {e}");
                }
                self.clone().sender(false, conf).await;
                seqno = ((seqno + 1) as u32).rem_euclid(pow) as u16;
                async_std::task::sleep(std::time::Duration::from_millis(
                    conf.mcast_hello_interval as u64 * 10,
                ))
                .await;
            }
        });
    }

    pub async fn ihu_timer(self, conf: Config, neighbors: Arc<Mutex<NeighborTable>>) {
        async_std::task::spawn(async move {
            loop {
                let neighbors = neighbors.lock_arc().await;
                let active_neighbors = neighbors.get_active_neighbors();
                for ne in &active_neighbors.neighbors {
                    let ae = if ne.is_ipv4() {
                        1
                    } else if ne.is_ipv6() {
                        if ne.to_string().starts_with("[fe80") {
                            3
                        } else {
                            2
                        }
                    } else {
                        0
                    };
                    let ihu = IHU {
                        ae,
                        rxcost: 30,
                        address: ne.ip(),
                        interval: conf.ihu_interval,
                    };
                    if let Err(e) = self.task_conn_tx.send(ihu.into_packet()).await {
                        log::error!("Error sending to thread connection: {e}");
                    }
                    self.clone().sender(false, conf).await;
                    async_std::task::sleep(std::time::Duration::from_millis(
                        conf.ihu_interval as u64 * 10,
                    ))
                    .await;
                }
            }
        });
    }

    pub async fn update_timer(self, conf: Config, prefixes: Advertise) {
        let mut seqno = 0;
        let interface = crate::nht::Nht::get_ifname_by_index(2).unwrap();
        loop {
            let addresses = crate::nht::Nht::get_ifaddrs(interface.clone()).unwrap();
            let mut update_buf = Vec::new();
            let router_id = RouterId {
                router_id: conf.router_id,
            };
            update_buf.append(&mut router_id.into_packet().packet_body);
            for prefix in &prefixes.prefix {
                let next_hop = crate::nht::Nht::get_primary_v4(addresses.clone()).unwrap_or(prefix.first_address());
                let ae = if prefix.is_ipv4() {
                    let next_hop = NextHop {
                        ae: 1,
                        next_hop,
                    };
                    update_buf.append(&mut next_hop.into_packet().packet_body);
                    1
                } else if prefix.is_ipv6() {
                    if let IpAddr::V6(addr) = prefix.address() {
                        if addr.segments()[0] == 0xFE80 {
                            3
                        } else {
                            2
                        }
                    } else {
                        0
                    }
                } else {
                    0
                };
                let update = Update {
                    ae,
                    flags: UpdateFlags::P,
                    plen: prefix.network_length(),
                    omitted: 0,
                    interval: conf.update_interval,
                    seqno,
                    metric: 0,
                    prefix: prefix.first_address(),
                };
                update_buf.append(&mut update.into_packet().packet_body);
                seqno += 1;
            }
            let packet = Packet {
                name: "Update",
                packet_body: update_buf,
                packet_trailer: None,
            };
            if let Err(e) = self.task_conn_tx.send(packet).await {
                log::error!("Error sending to thread connection: {e}");
            }
            self.clone().sender(false, conf).await;
            async_std::task::sleep(std::time::Duration::from_millis(
                conf.update_interval as u64 * 10,
            ))
            .await;
        }
    }
}
*/
