use std::net::IpAddr;

// Section 4.2: Packet Format
#[derive(Clone)]
pub struct Packet {
    pub name: &'static str,
    pub packet_body: Vec<u8>,
    pub packet_trailer: Option<Vec<u8>>,
}

impl Packet {
    pub fn to_bytes(self) -> Vec<u8> {
        let mut length = self.packet_body.len();
        let mut buf = Vec::new();
        buf.push(0x2A);
        buf.push(2);

        if let Some(trailer) = self.packet_trailer {
            length += trailer.len();
            let be_length = (length as u16).to_be_bytes();

            buf.push(be_length[0]);
            buf.push(be_length[1]);

            for b in self.packet_body {
                buf.push(b);
            }
            for b in trailer {
                buf.push(b);
            }
        } else {
            let be_length = (length as u16).to_be_bytes();

            buf.push(be_length[0]);
            buf.push(be_length[1]);

            for b in self.packet_body {
                buf.push(b);
            }
        };
        buf
    }
}

struct Tlv {
    r#type: u8,
    length: u8,
    payload: Vec<u8>,
}

impl Tlv {
    pub fn to_bytes(mut self) -> Vec<u8> {
        let mut buf = Vec::new();
        buf.push(self.r#type);
        buf.push(self.length);
        buf.append(&mut self.payload);
        buf
    }
}

pub trait IntoPacket {
    fn into_packet(self) -> Packet;
}

// Section 4.6: Details of Specific TLVs
// 4.6.1
struct Pad1();
impl IntoPacket for Pad1 {
    fn into_packet(self) -> Packet {
        Packet {
            name: "Pad1",
            packet_body: [0].to_vec(),
            packet_trailer: None,
        }
    }
}

// 4.6.2
struct PadN {
    length: u8,
}

impl IntoPacket for PadN {
    fn into_packet(self) -> Packet {
        let mbz = Vec::with_capacity(self.length as usize);
        let tlv = Tlv {
            r#type: 1,
            length: self.length,
            payload: mbz,
        };
        let tlv_bytes = tlv.to_bytes();
        Packet {
            name: "PadN",
            packet_body: tlv_bytes,
            packet_trailer: None,
        }
    }
}

// 4.6.3
pub struct AcknowledgementRequest {
    pub opaque: u16,
    pub interval: u16,
}

impl IntoPacket for AcknowledgementRequest {
    fn into_packet(self) -> Packet {
        let be_opaque = self.opaque.to_be_bytes();
        let be_interval = self.interval.to_be_bytes();
        let buf = [
            2,
            8,
            be_opaque[0],
            be_opaque[1],
            be_interval[0],
            be_interval[1],
        ];
        Packet {
            name: "AcknowledgementRequest",
            packet_body: buf.to_vec(),
            packet_trailer: None,
        }
    }
}

// 4.6.4
pub struct Acknowledgement {
    pub opaque: u16,
}

impl IntoPacket for Acknowledgement {
    fn into_packet(self) -> Packet {
        let be_opaque = self.opaque.to_be_bytes();
        let buf = [3, 4, be_opaque[0], be_opaque[1]];
        Packet {
            name: "Acknowledgement",
            packet_body: buf.to_vec(),
            packet_trailer: None,
        }
    }
}

// 4.6.5
#[derive(Debug)]
pub enum HelloFlags {
    Unicast,
    Multicast,
}

impl From<u16> for HelloFlags {
    fn from(value: u16) -> Self {
        match value {
            0x8000 => Self::Unicast,
            _ => Self::Multicast,
        }
    }
}

#[derive(Debug)]
pub struct Hello {
    pub flags: HelloFlags,
    pub seqno: u16,
    pub interval: u16,
}

impl IntoPacket for Hello {
    fn into_packet(self) -> Packet {
        let be_flags = match self.flags {
            HelloFlags::Unicast => 0x8000_u16,
            HelloFlags::Multicast => 0x0_u16,
        }
        .to_be_bytes();
        let be_seqno = self.seqno.to_be_bytes();
        let be_interval = self.interval.to_be_bytes();
        let buf = [
            4,
            6,
            be_flags[0],
            be_flags[1],
            be_seqno[0],
            be_seqno[1],
            be_interval[0],
            be_interval[1],
        ];
        Packet {
            name: "Hello",
            packet_body: buf.to_vec(),
            packet_trailer: None,
        }
    }
}

// 4.6.6
#[derive(Debug)]
pub struct IHU {
    pub ae: u8,
    pub rxcost: u16,
    pub interval: u16,
    pub address: IpAddr,
}

impl IHU {
    pub fn new(rxcost: u16, interval: u16, address: IpAddr) -> Self {
        let ae: u8 = if address.is_ipv4() {
            1
        } else if address.is_ipv6() {
            if address.to_string().starts_with("fe80") {
                3
            } else {
                2
            }
        } else {
            0
        };

        Self {
            ae,
            rxcost,
            interval,
            address,
        }
    }
}

impl IntoPacket for IHU {
    fn into_packet(self) -> Packet {
        let ae: u8 = if self.address.is_ipv4() {
            1
        } else if self.address.is_ipv6() {
            if self.address.to_string().starts_with("fe80") {
                3
            } else {
                2
            }
        } else {
            0
        };
        let rxcost = self.rxcost.to_be_bytes();
        let interval = self.rxcost.to_be_bytes();
        let addr_str = self.address.to_string();
        let address = addr_str.as_bytes();
        let length = 6 + address.len();
        let mut buf = Vec::with_capacity(2 + length);
        buf.push(5);
        buf.push(length as u8);
        buf.push(ae);
        buf.push(0);
        buf.push(rxcost[0]);
        buf.push(rxcost[1]);
        buf.push(interval[0]);
        buf.push(interval[1]);
        for b in address {
            buf.push(*b);
        }
        Packet {
            name: "IHU",
            packet_body: buf,
            packet_trailer: None,
        }
    }
}

// 4.6.7
pub struct RouterId {
    pub router_id: u64,
}

impl IntoPacket for RouterId {
    fn into_packet(self) -> Packet {
        let router_id = self.router_id.to_be_bytes();
        let buf = [
            6,
            10,
            0,
            0,
            router_id[0],
            router_id[1],
            router_id[2],
            router_id[3],
            router_id[4],
            router_id[5],
            router_id[6],
            router_id[7],
        ];
        Packet {
            name: "Router-id",
            packet_body: buf.to_vec(),
            packet_trailer: None,
        }
    }
}

// 4.6.8
pub struct NextHop {
    pub ae: u8,
    pub next_hop: IpAddr,
}

impl IntoPacket for NextHop {
    fn into_packet(self) -> Packet {
        let len = if self.next_hop.is_ipv4() { 6 } else { 18 };
        let mut buf = Vec::new();
        buf.push(7);
        buf.push(len);
        buf.push(self.ae);
        buf.push(0);
        match self.next_hop {
            IpAddr::V4(addr) => {
                for b in addr.octets() {
                    buf.push(b);
                }
            }
            IpAddr::V6(addr) => {
                for b in addr.octets() {
                    buf.push(b);
                }
            }
        }
        Packet {
            name: "NextHop",
            packet_body: buf,
            packet_trailer: None,
        }
    }
}

// 4.6.9
pub enum UpdateFlags {
    P,
    R,
}

pub struct Update {
    pub ae: u8,
    pub flags: UpdateFlags,
    pub plen: u8,
    pub omitted: u8,
    pub interval: u16,
    pub seqno: u16,
    pub metric: u16,
    pub prefix: IpAddr,
}

impl IntoPacket for Update {
    fn into_packet(self) -> Packet {
        let len = match self.prefix {
            IpAddr::V4(_addr) => 10 + self.plen.div_ceil(8),
            IpAddr::V6(_addr) => 10 + self.plen.div_ceil(4).div_ceil(2),
        };
        let interval = self.interval.to_be_bytes();
        let seqno = self.seqno.to_be_bytes();
        let metric = self.metric.to_be_bytes();
        let mut buf = Vec::new();
        buf.push(8);
        buf.push(len);
        buf.push(self.ae);
        match self.flags {
            UpdateFlags::P => buf.push(0x80),
            UpdateFlags::R => buf.push(0x40),
        }
        buf.push(self.plen);
        buf.push(self.omitted);
        buf.push(interval[0]);
        buf.push(interval[1]);
        buf.push(seqno[0]);
        buf.push(seqno[1]);
        buf.push(metric[0]);
        buf.push(metric[1]);

        match self.prefix {
            IpAddr::V4(addr) => {
                let octets = addr.octets();
                for i in 0..(len - 10) {
                    buf.push(octets[i as usize]);
                }
            }
            IpAddr::V6(addr) => {
                let octets = addr.octets();
                for i in 0..(len - 10) {
                    buf.push(octets[i as usize]);
                }
            }
        }

        Packet {
            name: "Update",
            packet_body: buf,
            packet_trailer: None,
        }
    }
}

// 4.6.10
struct RouteRequest {}

// 4.6.11
struct SeqnoRequest {}
