use crate::{
    nht::Nht,
    tables::neighbor::{NeighborTable, NeighborTableEntry, NeighborTableIndex},
    tlvs::{Acknowledgement, Hello, IntoPacket, Packet, IHU},
};
use cidr::{IpInet, Ipv6Inet};
use std::{
    net::{IpAddr, Ipv4Addr, Ipv6Addr},
    str::FromStr,
    sync::Arc,
};
use tokio::{net::UdpSocket, sync::Mutex};

static MAGIC: u8 = 42;
static VERSION: u8 = 2;

// UDP socket:
//  port: 6696
//  mcast: ff02::1:6
//  mcast: 224.0.0.111
#[derive(Clone)]
pub struct Socket {
    sock: Arc<UdpSocket>,
    tlv_queue: Vec<Packet>,
    own_ip: Ipv6Inet,
}

impl Socket {
    pub async fn new(iface_index: u32) -> anyhow::Result<Self> {
        let sock = Arc::new(UdpSocket::bind((Ipv6Addr::UNSPECIFIED, 6696)).await?);
        let tlv_queue = Vec::new();
        sock.join_multicast_v6(&Ipv6Addr::from_str("ff02::1:6").unwrap(), iface_index)?;
        let ifname = Nht::get_ifname_by_index(iface_index)?;
        let address = Nht::get_ifaddrs(ifname)?;

        let mut own_ip = None;
        for a in address {
            if let IpInet::V6(a) = a {
                if a.address().is_unicast_link_local() {
                    own_ip = Some(a);
                }
            }
        }

        if own_ip.is_none() {
            anyhow::bail!("Can't get primary link local for interface");
        };

        Ok(Self {
            sock,
            tlv_queue,
            own_ip: own_ip.unwrap(),
        })
    }

    /// Send packet to multicast
    pub async fn tx(&mut self) -> anyhow::Result<()> {
        let mut buf = Vec::new();
        for tlv in &self.tlv_queue {
            for b in &tlv.packet_body {
                buf.push(*b);
            }
        }

        if buf.len() == 0 {
            tracing::warn!("Attemted to send empty packet");
            return Ok(());
        }

        self.tlv_queue.clear();
        let packet = Packet {
            name: "",
            packet_body: buf,
            packet_trailer: None,
        }
        .to_bytes();
        tracing::trace!("Send packet: {packet:X?}");
        self.sock.send_to(&packet, "[ff02::1:6]:6696").await?;
        Ok(())
    }

    /// Queue packet for send to multicast
    pub async fn send(&mut self, packet: Packet) {
        tracing::trace!("Queuned packet: {}", packet.name);
        self.tlv_queue.push(packet);
    }

    pub async fn recv(&self, neighbors: Arc<Mutex<NeighborTable>>) -> anyhow::Result<()> {
        let mut buf = vec![0; 4];
        let (_count, peer) = self.sock.peek_from(&mut buf).await?;
        if peer.ip() == self.own_ip.first().address() {
            let _count = self.sock.recv(&mut buf).await?;
            tracing::trace!("Dropping own packet");
            return Ok(());
        }

        let magic = buf[0];
        let version = buf[1];

        if magic != MAGIC || version != VERSION {
            let _count = self.sock.recv(&mut buf).await?;
            anyhow::bail!(
                "Received invalid packet from {peer}: magic: {magic}, version: {version}"
            );
        }

        let length = u16::from_be_bytes([buf[2], buf[3]]);

        let mut buf = vec![0; length as usize + 4];
        let count = self.sock.recv(&mut buf).await?;

        tracing::trace!("Read {count} bytes {:X?} from {peer}", buf);
        let mut neighbors = neighbors.lock().await;
        let mut prev_v6 = None;

        let mut head: usize = 4;
        loop {
            if head >= length as usize {
                // The entire buffer has been read
                break Ok(());
            }

            let r#type = buf[head + 0];

            // Pad
            if r#type == 0 {
                tracing::trace!("Received pad");
                head += 1;
                continue;
            }

            let tlv_length = buf[head + 1];

            match r#type {
                // PadN
                1 => continue,
                // Acknowledgement Request
                2 => {
                    let opaque = u16::from_be_bytes([buf[head + 4], buf[head + 5]]);
                    let interval = u16::from_be_bytes([buf[head + 6], buf[head + 7]]);
                    tracing::trace!(
                        "Received Acknowledgement Request: opaque {opaque}, interval {interval}"
                    );

                    let acknowledgement = Acknowledgement { opaque };
                    self.sock
                        .send_to(&acknowledgement.into_packet().to_bytes(), peer)
                        .await?;
                }
                // Acknowledgement
                // TODO: process
                3 => {
                    let opaque = u16::from_be_bytes([buf[head + 2], buf[head + 3]]);
                    tracing::trace!("Received Acknowledgement: opaque: {opaque}");
                }
                // Hello
                4 => {
                    let flags = u16::from_be_bytes([buf[head + 2], buf[head + 3]]);
                    let seqno = u16::from_be_bytes([buf[head + 4], buf[head + 5]]);
                    let interval = u16::from_be_bytes([buf[head + 6], buf[head + 7]]);
                    let hello = Hello {
                        flags: flags.into(),
                        interval,
                        seqno,
                    };

                    tracing::trace!("Received Hello: {hello:?}");

                    // Add Hello to neighbor table
                    let index = NeighborTableIndex::parse(peer.to_string())?;
                    let neighbor = neighbors.get_neighbor(&index);
                    if let Some(neighbor) = neighbor {
                        neighbor.add_hello(hello).await?;
                    } else {
                        let neighbor = NeighborTableEntry::new().await;
                        neighbors.add_neighbor(index, neighbor);
                    }
                }
                // IHU
                5 => {
                    let ae = buf[head + 2];
                    let rxcost = u16::from_be_bytes([buf[head + 4], buf[head + 5]]);
                    let interval = u16::from_be_bytes([buf[head + 6], buf[head + 7]]);
                    let address = &buf[head + 8..head + (tlv_length + 2) as usize];
                    let address = Self::address_from_ae(ae, address, None, None)?;

                    let ihu = IHU {
                        ae,
                        rxcost,
                        interval,
                        address,
                    };

                    tracing::trace!("Received IHU: {ihu:?}");

                    // Update IHU timer for peer
                    let index = NeighborTableIndex::parse(peer.to_string())?;
                    let neighbor = neighbors.get_neighbor(&index);
                    if let Some(neighbor) = neighbor {
                        neighbor.add_ihu(ihu).await?;
                    } else {
                        tracing::warn!("Received Update from unknown neighbor: {:?}", peer);
                    }
                }
                // Router-Id
                // TODO: this TLV should alway be followed by an Update
                6 => {
                    let router_id = u64::from_be_bytes([
                        buf[head + 4],
                        buf[head + 5],
                        buf[head + 6],
                        buf[head + 7],
                        buf[head + 8],
                        buf[head + 9],
                        buf[head + 10],
                        buf[head + 11],
                    ]);
                    tracing::trace!("Received Router ID: router id: {router_id}");
                }
                // Next Hop
                // TODO: relevant for Update TLVs
                7 => {
                    let ae = buf[head + 2];
                    let addr_buf = &buf[head + 4..head + (tlv_length + 2) as usize];
                    let next_hop = Self::address_from_ae(ae, addr_buf, prev_v6, None)?;
                    tracing::trace!("Received Next Hop: ae: {ae}, next hop: {next_hop}");
                }
                // Update
                // TODO: process
                8 => {
                    let ae = buf[head + 2];
                    let flags = buf[head + 3];
                    let plen = buf[head + 4];
                    let ommitted = buf[head + 5];
                    let interval = u16::from_be_bytes([buf[head + 6], buf[head + 7]]);
                    let seqno = u16::from_be_bytes([buf[head + 8], buf[head + 9]]);
                    let metric = u16::from_be_bytes([buf[head + 10], buf[head + 11]]);
                    let prefix_buf = &buf[head + 12..head + (tlv_length + 2) as usize];
                    let prefix = Self::address_from_ae(ae, prefix_buf, prev_v6, Some(plen))?;
                    
                    if let IpAddr::V6(addr) = prefix {
                        prev_v6 = Some(addr);
                    }

                    tracing::trace!("Received Update: ae: {ae}, flags: {flags}, plen: {plen}, ommitted: {ommitted}, interval: {interval}, seqno: {seqno}, metric: {metric}, prefix: {prefix}");
                }
                // Route request
                // TODO: Respond
                9 => {
                    let ae = buf[head + 2];
                    let plen = buf[head + 3];
                    let prefix_buf = &buf[head + 4..head + (tlv_length + 2) as usize];
                    let prefix = Self::address_from_ae(ae, prefix_buf, prev_v6, Some(plen))?;
                    tracing::trace!(
                        "Received Route Request: ae: {ae}, plen: {plen}, prefix: {prefix}"
                    );
                }
                // Seqno request
                // TODO: Respond
                10 => {
                    let ae = buf[head + 2];
                    let plen = buf[head + 3];
                    let seqno = u16::from_be_bytes([buf[head + 4], buf[head + 5]]);
                    let hop_count = buf[head + 6];
                    let router_id = u64::from_be_bytes([
                        buf[head + 8],
                        buf[head + 9],
                        buf[head + 10],
                        buf[head + 11],
                        buf[head + 12],
                        buf[head + 13],
                        buf[head + 14],
                        buf[head + 15],
                    ]);
                    let prefix_buf = &buf[head + 16..head + (tlv_length + 2) as usize];
                    let prefix = Self::address_from_ae(ae, prefix_buf, prev_v6, Some(plen))?;
                    tracing::trace!("Received Seqno Request: ae: {ae}, plen: {plen}, seqno: {seqno}, hop count: {hop_count}, router id: {router_id}, prefix: {prefix}");
                }
                _ => tracing::warn!("Received unknown TLV type: {type}"),
            }

            head += tlv_length as usize + 2;
        }
    }

    fn address_from_ae(
        ae: u8,
        address: &[u8],
        prev_v6: Option<Ipv6Addr>,
        plen: Option<u8>,
    ) -> anyhow::Result<IpAddr> {
        Ok(match ae {
            0 => {
                IpAddr::from_str("0.0.0.0")?
            }
            1 => IpAddr::V4(Ipv4Addr::new(
                address[0], address[1], address[2], address[3],
            )),
            2 => {
                if let Some(addr) = prev_v6 {
                    if plen.is_none() {
                        anyhow::bail!("No plen");
                    } 
                    let plen = plen.unwrap();
                    let discard_len = ((128 - plen) / 8) as usize;
                    let cutoff = if discard_len < address.len() {
                        address.len()
                    } else {
                        discard_len + address.len()
                    };

                    let octets = addr.octets();
                    let mut buf = Vec::new();

                    for b in &octets[0..(16 - cutoff)] {
                        buf.push(*b);
                    }

                    for b in address {
                        buf.push(*b);
                    }

                    if cutoff > address.len() {
                        let pad = cutoff - address.len();
                        for _ in 0..pad {
                            buf.push(0);
                        }
                    }
                    
                    IpAddr::V6(Ipv6Addr::from([
                        buf[0],
                        buf[1],
                        buf[2],
                        buf[3],
                        buf[4],
                        buf[5],
                        buf[6],
                        buf[7],
                        buf[8],
                        buf[9],
                        buf[10],
                        buf[11],
                        buf[12],
                        buf[13],
                        buf[14],
                        buf[15],
                    ]))
                } else {
                    if address.len() != 16 {
                        anyhow::bail!("Invalid IPv6 address length");
                    }

                    IpAddr::V6(Ipv6Addr::from([
                        address[0],
                        address[1],
                        address[2],
                        address[3],
                        address[4],
                        address[5],
                        address[6],
                        address[7],
                        address[8],
                        address[9],
                        address[10],
                        address[11],
                        address[12],
                        address[13],
                        address[14],
                        address[15],
                    ]))
                }
            }
            3 => IpAddr::V6(Ipv6Addr::from([
                0xFE, 0x80, 0, 0, 0, 0, 0, 0, address[7], address[6], address[5], address[4],
                address[3], address[2], address[1], address[0],
            ])),
            _ => {
                anyhow::bail!("Unknown address encoding: {ae} {address:?}");
            }
        })
    }
}
